#!/usr/bin/perl -w
#
###################################################
#$Version: 1.0$
#$Author: Daniel Koffler$
#
# NXE - Nagios XML Engine for Nagios 1.x
#
# This application provides several methods of reporting data from Nagios in XML format
# as well as transforming that XML via XSLT. An XML based command interpreter is also included.
#
# This application can be accessed via XML-RPC or HTTP Get when installed as an Apache
# or IIS (or any other Perl capable Web Server) CGI. Alternatively, you can run this
# application from the command line.
#
# Please see the README file that came with this application for more details.
#
# Nagios is a trademark of Ethan Galstad.
# For more information see http://www.nagios.org or http://nxe.sourceforge.net
#
# Copyright (C) 2003 Daniel Koffler
#
#This library is free software; you can redistribute it and/or
#modify it under the terms of the GNU Lesser General Public
#License as published by the Free Software Foundation; either
#version 2.1 of the License, or (at your option) any later version.
#
#This library is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#Lesser General Public License for more details.
#
#You should have received a copy of the GNU Lesser General Public
#License along with this library; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
####################################################
# Revision History:
# Revision 1.0 June 2003 Daniel Koffler
####################################################
#      ADDITIONAL INSTRUCTIONS:
#
# Modify the $CONFIG_FILE variable below to reflect the path to the nxe.config file.
# Please see the INSTALL file that came with this application for details.
####################################################


my $CONFIG_FILE = "nxe.config";
my @xslt_procs = ("libxslt", "sablotron", "xalan", "xslt", "xt", "saxon");

use strict;
use XML::Writer;
use XML::Writer::String;
use Sys::Hostname;
use Frontier::RPC2;
use XML::XSLT::Wrapper;
use XML::Sablotron;
use CGI qw(:standard escapeHTML);

my $cmd_type;
my %files;
my $rpcsrv = Frontier::RPC2->new() if ($ENV{"REQUEST_METHOD"} eq "POST"); #setup RPC Server if required
my $rpc_string;
my $xhelp = XML::Writer::String->new(); 
my $wr = new XML::Writer( DATA_MODE => 'true', DATA_INDENT => 2, OUTPUT => $xhelp);
$wr->xmlDecl("UTF-8");

sub bail {
    #my $errornum = shift @_;
    my $errornum = "900";
    my $error = "@_";
    my $rpc_string = "";
    if ($ENV{"REQUEST_METHOD"} eq "POST"){ #XML-RPC Request
	$rpc_string = $rpcsrv->encode_fault("$errornum", $error);
	my $length = length($rpc_string);
	print <<"EOD";
Status: 200 OK
Content-type: text/xml
Cache-control: no-cache
Pragma: no-cache
Content-length: $length

EOD
	print $rpc_string;
	die "Error: $error";
	
    }else{ #Is in GET or command line mode
	print "Content-type: text/xml","\n\n" if ($ENV{"REQUEST_METHOD"} eq "GET"); #if get return header
	my $xhelp = XML::Writer::String->new(); 
	my $wr = new XML::Writer( DATA_MODE => 'true', DATA_INDENT => 2, OUTPUT => $xhelp);
	$wr->xmlDecl("UTF-8");
	$wr->doctype("NXE_RESULT", "-//RSMEXPERTS//DTD Nagios XML Engine Output//EN//", "http://nxe.sourceforge.net/dtd/NXEv1r0.dtd");
	$wr->startTag("NXE_RESULT", "NAME" => "Nagios XML Engine Report");
	$wr->startTag("ERROR", "CODE" => "$errornum");
	$wr->characters("$error");
	$wr->endTag();
	$wr->endTag();
	$wr->end();
	if ($cmd_type eq "transform" | $cmd_type eq "cmd_transform"){ #Transform error message
	    
	    my $xsl_string;
	    my $xslt = XML::XSLT::Wrapper->new(
	    ProcessorList => \@xslt_procs);
	    my $cgi_params = param('params');
	    my @params = split(/:/,$cgi_params);
	    my $tmp = $/;
	    $/ = undef;
	    open(XSLT, $files{$params[0]}) || (print $xhelp->value() && die "Error: $error");
	    $xsl_string = <XSLT>;
	    $/ = $tmp;
	    close(XSLT);
	    $rpc_string = $xslt -> transform(XSLString => $xsl_string,
				     XMLString => $xhelp->value) || (print $xhelp->value() && die "Error: $error");
	    print $rpc_string;
	}else{ #Print xml error value
	    print $xhelp -> value();
	}
	die "Error: $!";
    }
}

sub status {
    # Return Status
    $cmd_type = "status";
    my @targets = "@_";
    $wr->doctype("NXE_RESULT", "-//RSMEXPERTS//DTD Nagios XML Engine Output//EN//", "http://xmlexperts.org/nxe/dtd/v1r0");	
    my $nagver;
    my %services;
    my %hosts;
    my @nagproc;

    $wr->startTag("NXE_RESULT",
	    "NAME" => "Nagios XML Engine Report",
	    "SOURCE" => "NXE");
    foreach my $target (@targets){
	
	open(STATUS, $files{$target}) || bail "Can't open Nagios Status File $target. Check config file.";
	while (defined (my $line = <STATUS>)){
	    chomp($line);
	    if ($line =~ s/^# //){
		$nagver = $line;
		next;
	    }
	    my @ent = split(/;/,$line);
	    my @type = ($ent[0] =~ /^(\[)(\d+)(\]\s)(\w+)/) or bail "Invalid format in the following entry:\n $line";
	    shift(@ent);    
	    if ($type[3] eq "PROGRAM"){
		@nagproc = ($type[1],@ent);
	    }elsif ($type[3] eq "HOST"){
		@hosts{$ent[0]} = [$type[1], @ent];
	    }elsif ($type[3] eq "SERVICE"){
		@services{"$ent[0] - $ent[1]"} = [$type[1], @ent];
	    }else {bail "Unkown Entry Type"}    
	}
	
	close(STATUS) || bail "Filehandle to $target was interupted";
	
	$wr->startTag("NAG_PROC",
	    "SOURCE" => $target,
	    "LAST_UPDATE" => $nagproc[0],
	    "VERSION" => $nagver,
	    "HOST_NAME" => hostname,
	    "P_START_TIME" => $nagproc[1],
	    "P_PID" => $nagproc[2],
	    "DAEMON" => $nagproc[3],
	    "P_LAST_CHECK" => $nagproc[4],
	    "LOG_ROTATION" => $nagproc[5],
	    "NOTIFIC" => $nagproc[6],
	    "EXEC_SVC_CHK" => $nagproc[7],
	    "ACEP_PASV_CHK" => $nagproc[8],
	    "EVENT_HNDLR" => $nagproc[9],
	    "OBSS_SVC" => $nagproc[10],
	    "FLAP_DETECT" => $nagproc[11],
	    "FAIL_PREDICT" => $nagproc[12],
	    "PERF_DATA" => $nagproc[13]);
	
	my @hostkeys = sort( keys(%hosts));	
	foreach my $key (@hostkeys){
	    my @host = @{$hosts{$key}};
	    $wr -> emptyTag("HOST",
		"LAST_UPDATE" => $host[0],
		"HOSTNAME" => $host[1],
		"HST_STATUS" => $host[2], 
		"LAST_CHECK" => $host[3],
		"LAST_STATE_CHANGE" => $host[4],
		"HST_ACK" => $host[5],
		"TIME_UP" => $host[6],
		"TIME_DOWN" => $host[7],
		"TIME_UNREACH" => $host[8],
		"LAST_NOTIFIC" => $host[9],
		"CURR_NOTF_NUM" => $host[10],
		"NOTF_EN" => $host[11],
		"EVENT_HNDLR_EN" => $host[12],
		"CHECKS_EN" => $host[13],
		"FLAP_EN" => $host[14],
		"FAIL_PREDICT_EN" => $host[15],
		"PERF_DATA" => $host[16],
		"HST_FLAPING" => $host[17],
		"PERC_STATE_CHNG" => $host[18],
		"SCHD_DWTM_DEPTH" => $host[19],
		"HST_PLUG_OUTPUT" => $host[20]);
	}
	my @srvkeys = sort( keys(%services));
	foreach my $key (@srvkeys){
	    my @service = @{$services{$key}};
	    $wr->emptyTag("SERVICE",
		"LAST_UPDATE" => $service[0],
		"HOSTNAME" => $service[1],
		"SVC_DESC" => $service[2],
		"SVC_STATUS" => $service[3],
		"LAST_HSTATE" => $service[4],
		"RETRY" => $service[5],
		"STATE_TYPE" => $service[6],
		"LAST_CHECK" => $service[7],
		"NEXT_CHECK" => $service[8],
		"CHECK_TYPE" => $service[9],
		"LAST_STATE_CHANGE" => $service[10],
		"CHECKS_EN" => $service[11],
		"PASV_EN" => $service[12],
		"EVENT_HNDLR_EN" => $service[13],
		"SVC_ACK" => $service[14],
		"TIME_UP" => $service[15],
		"TIME_UNKWN" => $service[16],
		"TIME_WARN" => $service[17],
		"TIME_CRIT" => $service[18],
		"LAST_NOTIFIC" => $service[19],
		"CURR_NOTF_NUM" => $service[20],
		"NOTF_EN" => $service[21],
		"LATENCY" => $service[22],
		"EXEC_TIME" => $service[23],
		"FLAP_EN" => $service[24],
		"SVC_FLAPING" => $service[25],
		"FAIL_PREDICT_EN" => $service[26],
		"PERC_STATE_CHNG" => $service[27],
		"SCHD_DWTM_DEPTH" => $service[28],
		"PERF_DATA" => $service[29],
		"SVC_OBSS_EN"  => $service[30],
		"SVC_PLUG_OUTPUT" => $service[31]);
	}
    }
    $wr->endTag();
    $wr->endTag();
    $wr->end();
    return ($xhelp -> value());
}

sub cmd {# Create a Nagios external command entry
    $cmd_type = "cmd";
    my $target = shift @_;
    my $cmdstr = join(";",@_);
    my $cmdtime = time;
    $cmdstr = "[$cmdtime] $cmdstr\n";
    open (CMDFILE, ">>$files{$target}") || bail "Unable to open file $target for writing.";
    print CMDFILE $cmdstr;
    close (CMDFILE);
    $wr->doctype("NXE_RESULT", "-//XMLEXPERTS//DTD Nagios XML Engine Output//EN//", "http://nxe.sourceforge.net/dtd/NXEv1r0.dtd");	
    $wr->startTag("NXE_RESULT", "NAME" => "Nagios XML Engine Report",
		    "SOURCE" => "$target");
    $wr->startTag("CMD_RESULT", "TIME" => $cmdtime,
		    "TARGET" => "$target",
		    "STATUS" => "success");
    $wr->characters("$cmdstr");
    $wr->endTag();
    $wr->endTag();
    $wr->end();
    $rpc_string = $xhelp->value();
    
}

sub transform { #do a transform
    $cmd_type = "transform";
    my $xsl_string;
    my $xslt = XML::XSLT::Wrapper->new(
		ProcessorList => \@xslt_procs);
    my $xsl_alias = shift @_;
    my @targets = @_;
    my $tmp = $/;
    $/ = undef;
    open(XSLT, $files{$xsl_alias}) || ($cmd_type="" && bail "Unable to open XSL file $xsl_alias.");
    $xsl_string = <XSLT>;
    $/ = $tmp;
    #$xsl_string =~ s/\n//g;
    close(XSLT);
    my $result = status(@targets);
    #$result =~ s/\n//g;
    #my $sab = new XML::Sablotron();
    #XML::Sablotron::ProcessStrings($xsl_string, $result, $rpc_string);
    
    $rpc_string = $xslt -> transform(XSLString => $xsl_string,
    				     XMLString => $result) || bail "Files OK. Transformation Engine Failure.\n
    				    Make sure you have installed a compatible XSLT Processor.\nError: $!";
    return ($rpc_string);
}

sub transform_cmd { #transform results of a command
    my $xsl_file = shift @_;
    $rpc_string = cmd(@_);
    my $xsl_string;
    my $xslt = XML::XSLT::Wrapper->new(
    ProcessorList => \@xslt_procs);
    my $tmp = $/;
    $/ = undef;
    open(XSLT, $files{$xsl_file}) ||  die "Can't open file $xsl_file";
    $xsl_string = <XSLT>;
    $/ = $tmp;
    close(XSLT);
    $rpc_string = $xslt -> transform(XSLString => $xsl_string,
			XMLString => $rpc_string) || die "Unable to transform command result";
    return ($rpc_string);
}

#main logic
open (CONFIG, $CONFIG_FILE) || bail "Could not open NSX configuration file $CONFIG_FILE.\nError: $!";
while (<CONFIG>) {
    chomp;                  # no newline
    s/#.*//;                # no comments
    s/^\s+//;               # no leading white
    s/\s+$//;               # no trailing white
    next unless length;     # anything left?
    my ($var, $value) = split(/\s*=\s*/, $_, 2);
    $files{$var} = $value;
} 


if ($ENV{"REQUEST_METHOD"} eq "POST"){ #XML-RPC Mode
    my $length = $ENV{'CONTENT_LENGTH'}; 
    my $rpc_req;
    my $count = read STDIN, $rpc_req, $length; #Check to make sure we got the whole request
    bail "XML-RPC Request content didn't match expected length. You tried to send $length but I got $count." unless $count == $length;
    my $methods = {
	cmd => \&cmd,
	status => \&status,
	transform => \&transform,
	transform_cmd => \&transform_cmd
    };
    $rpc_string = $rpcsrv->serve($rpc_req, $methods);
    $length = length($rpc_string);
    print <<"EOD";
Status: 200 OK
Content-type: text/xml
Cache-control: no-cache
Pragma: no-cache
Content-length: $length

EOD
    print $rpc_string;
    exit(0);
    
}elsif ($ENV{"REQUEST_METHOD"} eq "GET") { #CGI Mode
    my $length;
    my $type = param('type');
    my $cgi_params = param('params');
    my @params = split(/:/,$cgi_params);
    if ($type eq "status"){
	$rpc_string = status(@params);
	$length = length($rpc_string);
	print <<"EOD";
Status: 200 OK
Content-type: text/xml
Cache-control: no-cache
Pragma: no-cache
Content-length: $length

EOD
	print $rpc_string;
	exit(0);
	
    }elsif ($type eq "cmd"){
	$rpc_string = cmd(@params);
	$length = length($rpc_string);
	print <<"EOD";
Status: 200 OK
Content-type: text/xml
Cache-control: no-cache
Pragma: no-cache
Content-length: $length

EOD
	print $rpc_string;
	exit(0);
    }elsif ($type eq "transform"){
	my $cont_type = shift @params;
	bail "You must specify an output type of HTML or XML for HTTP GET transform commands." unless ($cont_type eq "xml" || $cont_type eq "html");
	$rpc_string = transform(@params);
	$length = length($rpc_string);
	print <<"EOD";
Status: 200 OK
Content-type: text/$cont_type
Cache-control: no-cache
Pragma: no-cache
Content-length: $length

EOD
	print $rpc_string;
	exit(0);
    }elsif ($type eq "transform_cmd"){
	 my $cont_type = shift @params;
	         bail "You must specify an output type of HTML or XML for HTTP GET transform commands." unless ($cont_type eq "xml" || $cont_type eq "html");
	$rpc_string = transform_cmd(@params);
	$length = length($rpc_string);
	print <<"EOD";
Status: 200 OK
Content-type: text/$cont_type
Cache-control: no-cache
Pragma: no-cache
Content-length: $length

EOD
	print $rpc_string;
	exit(0);
	
    }else{
	bail "Unknown Command Type: $type.\nUse: http://localhost/nsx?type=status&params=probe1;probe2;probe3\n
	type = status | cmd | transform\n
	params = FILE_ALIAS1;FILE_ALIAS2;FILE_ALIAS3 as defined in the configuration file."
    }


}else { #Command line mode
    my $type = shift @ARGV;
    my @params = @ARGV;
    if ($type eq "status"){#Status
	print status(@params) || bail "Unable to retrieve status";
	exit(0);
    }elsif ($type eq "cmd"){ #Command
	print cmd(@params) || bail "Unable to perform command";
	exit(0);
    }elsif ($type eq "transform"){#Transform
	print transform(@params) || bail "Unable to transform status";
	exit(0);
    }elsif ($type eq "transform_cmd"){#Transform command output
	print transform_cmd(@params) || bail "Unable to transform command";
	exit(0)
    }else {bail "Unkown command type: $type."}
    
}
